/*
** Module   : BrowseColumnWidget.java
** Abstract : server side browse column widget implementation
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 NVS 20051026   @23480 Created initial version.
** 002 SIY 20060227   @25122 Changes for the browse controller 
**                           implementation.
** 003 SIY 20060330   @25278 Added methods for LABEL attribute.
** 004 SIY 20060417   @26218 Added setColumnLabel() method.
** 005 GES 20060803   @28365 Major cleanup and refactoring.
** 006 GES 20080320   @37580 Added TYPE attribute.
** 007 SIY 20080603   @38474 setColumn() now operates on separate
**                           COLUMN-LABEL attribute. 
** 008 SIY 20090206   @41262 Changed setLabel() logic to make it properly
**                           handling setLabel/setColumnLabel/setLavel 
**                           sequences (this can happen when label is
**                           dynamically set after browse construction).
** 009 SIY 20090218   @41328 Added support for AT base field format width.
** 010 SIY 20111101          Updated imports.
** 011 CA  20130221          Refactored usage of PRIVATE-DATA and TYPE attribute.
** 012 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 013 CA  20130927          Added LegacyResource annotation. TODO: implement the resource type 
**                           based actual column type and "-browcoltype" parameter.
** 014 CA  20131115          Changes related to properly instantiate dynamic resources.
** 015 CA  20140711          Enhanced font support, for GUI clients.
** 016 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 017 CA  20141104          Renamed setWidth to setWidthChars.
** 018 HC  20141021          Implemented window move/position support.
** 019 SVL 20141102          Added some attributes.
** 020 SVL 20141209          Implemented getNextColumn.
** 021 GES 20141124          Added inner-lines support. Added several combo-box attrs/meths. 
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Server side representation of a browse column.
 * <p>
 * The {@link LegacyResource} annotation targets the {@link LegacyResource#FILL_IN} type.
 * <p>
 * The 4GL documentation states that {@link LegacyResource#BROWSE_COLUMN} type is used only when 
 * "-browcoltype" parameter is * at program startup.  Else, the type can be either 
 * {@link LegacyResource#FILL_IN}, {@link LegacyResource#COMBO_BOX} or 
 * {@link LegacyResource#TOGGLE_BOX} (depending on the column's widget).
 * 
 * TODO: fix this to report the correct type. 
 * TODO: check if the browse columns can be dynamic; if so, change the c'tor to receive a boolean
 *       flag.
 */
@LegacyResource(resource = LegacyResource.BROWSE_COLUMN)
public class BrowseColumnWidget
extends GenericWidget<BrowseColumnConfig>
implements AutoZapElement,
           BrowseColumnInterface,
           BrowseElement,
           MaxCharsElement,
           NumItems,
           SortedElement,
           InnerLines
{
   /** Back reference to the variable accessor. */
   private FrameElement element = null;

   /**
    * Default constructor
    */
   public BrowseColumnWidget()
   {
      super(false, new BrowseColumnConfig());
   }

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   public logical isReadOnly() 
   {
      return new logical(config.readOnly);
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(logical r) 
   {
      config.readOnly = r.booleanValue();
      pushScreenDefinition();
   }
   
   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   public integer getInnerLines()
   {
      return new integer(config.innerLines);
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(double lines)
   {
      config.innerLines = (int) lines;
      pushScreenDefinition();
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         UnimplementedFeature.todo("INNER-LINES attribute unknown behavior is not known.");
         return;
      }
   
      setInnerLines(lines.intValue());
   }
   
   /**
    * Gets the LIST-ITEMS writeable attribute.
    * 
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItems()
   {
      UnimplementedFeature.missing("LIST-ITEMS attribute getter");
      return new character();
   }

   /**
    * Sets the LIST-ITEMS writeable attribute.
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItems(character list)
   {
      UnimplementedFeature.missing("LIST-ITEMS attribute setter");
   }
   
   /**
    * Setter for the LIST-ITEMS writeable attribute.
    *
    * @param    list
    *           The new value for the LIST-ITEMS.
    */
   public void setListItems(String list)
   {
      UnimplementedFeature.missing("LIST-ITEMS attribute setter");
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character itemList)
   {
      if (itemList == null || itemList.isUnknown())
      {
         UnimplementedFeature.todo("ADD-LAST method unknown behavior is not known");
         return new logical();
      }
      
      return addLast(itemList.toStringMessage());
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String itemList)
   {
      UnimplementedFeature.missing("ADD-LAST method");
      return new logical();
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, BaseDataType value)
   {
      if (label == null || label.isUnknown() || value == null || value.isUnknown())
      {
         UnimplementedFeature.todo("ADD-LAST method unknown behavior is not known");
         return new logical();
      }
      
      UnimplementedFeature.missing("ADD-LAST method");
      return new logical();
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a single label
    * and value pair to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String label, BaseDataType value)
   {
      return addLast(new character(label), value);
   }
   
   /**
    * Implements the DELETE() widget method, which removes the list item or items specified.
    *
    * @param    item
    *           The item or items to delete.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character item)
   {
      UnimplementedFeature.missing("DELETE() method");
      return new logical();
   }
   
   /**
    * Implements the DELETE() widget method, which removes the list item or items specified.
    *
    * @param    item
    *           The item or items to delete.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(String item)
   {
      UnimplementedFeature.missing("DELETE() method");
      return new logical();
   }
   
   /**
    * Implements the DELETE() widget method, which removes the list item at the given index.
    *
    * @param    idx
    *           The index of the item to delete.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(int64 idx)
   {
      UnimplementedFeature.missing("DELETE() method");
      return new logical();
   }
   
   /**
    * Implements the DELETE() widget method, which removes the list item at the given index.
    *
    * @param    idx
    *           The index of the item to delete.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(long idx)
   {
      UnimplementedFeature.missing("DELETE() method");
      return new logical();
   }
   
   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   public logical isModified()
   {
      // TODO: 4GL docs suggest we need to realize the widget when this is called
      // TODO: 4GL docs state this attribute is read-only for browse column
      UnimplementedFeature.partial("MODIFIED attribute runtime behavior is incomplete");
      return new logical(config.modified);
   }
   
   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn)
   {
      config.autoReturn = autoReturn;
      pushScreenDefinition();
   }
   
   /**
    * Set the COLUMN-LABEL attribute.
    * 
    * @param    label
    *           The new value for the attribute.
    */
   public void setColumnLabel(String label)
   {
      config.clabel = label;
      pushScreenDefinition();
   }
   
   /**
    * Sets the label text, removing any '&' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   public void setLabel(String label)
   {
      // TODO: why is this not removing any "&" char first, as column-label ends up with with them,
      // if they exist?

      if (config.label != null)
         config.clabel = label;
      
      super.setLabel(label);
   }

   /**
    * Set the DATA-TYPE attribute value.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setDataType(String value)
   {
      config.dataType = value;
      pushScreenDefinition();
   }

   /**
    * Queries the DATA-TYPE attribute of this widget.
    *
    * @return A character value that represents the data type of the field
    *         associated with the widget.
    */
   @Override
   public character getDataType()
   {
      return new character(config.dataType);
   }

   /**
    * Sets new value of DATA-TYPE attribute.
    *
    * @param dataType
    *        new value of DATA-TYPE attribute
    */
   @Override
   public void setDataType(character dataType)
   {
      // TODO: implement this
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Set the WIDTH attribute value.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setWidthChars(double value)
   {
      config.autoResize = false;
      config.widthChars = Coordinate.scale(value);
      pushScreenDefinition();
   }

   /**
    * Set format width of AT base field.
    * 
    * @param   value
    *          New format width.
    */
   public void setFormatWidth(int value)
   {
      config.formatWidth = value;
      // We do not push definition here because
      // this is not a run-time changeable value, 
      // it set only once during startup.
   }
   
   /**
    * Enable editing of the column.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setEditable(boolean value)
   {
      config.readOnly = !value;
      pushScreenDefinition();
   }

   /**
    * Set AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    * 
    * @param    value
    *           New value for the attribute.
    */
   public void setAutoZap(boolean value)
   {
      // TODO is that correct? We should correlate this with AUTO-ZAP attribute of inner cells.
      config.autoZap = value;
      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(boolean labels)
   {
      config.labels = labels;
      if (config.autoResize)   // trigger size recalculation.
         config.widthChars = -1;
      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    nolabels
    *           New value for the attribute.
    */
   public void setNoLabels(boolean nolabels)
   {
      setLabels(!nolabels);
   }
   
   /**
    * Force presence of the label.
    * 
    * @param    force
    *           New value for the flag.
    */
   public void setForceLabel(boolean force)
   {
      config.forceLabel = force;
      if (config.autoResize)   // trigger size recalculation.
         config.widthChars = -1;
      pushScreenDefinition();
   }
   
   /**
    * Provide link to the backing accessor.
    * 
    * @param    element
    *           The accessor for this column.
    */
   protected void setElement(FrameElement element)
   {
      this.element = element;
   }
   
   /**
    * Retrieve value of the associated variable.
    * 
    * @return   Current value of the variable.
    */
   public BaseDataType getValue()
   {
      return element.get();
   }

   /**
    * Assign value to associated variable.
    * 
    * @param    value
    *           New value for the variable.
    */
   public void setValue(BaseDataType value)
   {
      element.set(value);
   }

   /**
    * Set ID of the browse widget to which this column belongs.
    * 
    * @param    id
    *           Browse ID.
    */
   public void setBrowseId(int id)
   {
      config.browseId = id;
   }

   /**
    * Get the COLUMN-FONT attribute of this browse column widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getColumnFont()
   {
      int font = config.columnFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the COLUMN-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setColumnFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.columnFont = font;
      pushScreenDefinition();
   }

   /**
    * Get the LABEL-FONT attribute of this browse column widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getLabelFont()
   {
      int font = config.labelFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the LABEL-FONT attribute of this browse widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setLabelFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.labelFont = font;
      pushScreenDefinition();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   public void setPfColor(NumberType param)
   {
      config.columnPfColor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   public void setPfColor(Color inputColor)
   {
      config.columnPfColor = inputColor;
      pushScreenDefinition();
   }
   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   public void setDColor(NumberType param)
   {
      config.columnDColor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDColor(Color displayColor)
   {
      config.columnDColor = displayColor;
      pushScreenDefinition();
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-BGCOLOR attribute.
    */
   public void setBgColor(NumberType param)
   {
      config.columnBgColor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-BGCOLOR attribute.
    */
   public void setBgColor(Color param)
   {
      config.columnBgColor = param;
      pushScreenDefinition();
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    param
    *           The new value for the COLUMN-BGCOLOR attribute.
    */
   public void setBgColor(int param)
   {
      config.columnBgColor = new ColorSpec(param).convert();
      pushScreenDefinition();
   }

   /**
    * Get the NUM-ITEMS attribute.
    *
    * @return   The numebr of items in the list.
    */
   public integer getNumItems()
   {
      UnimplementedFeature.missing("NUM-ITEMS attribute getter");
      return new integer();
   }
   
   /**
    * Implements the NEXT-COLUMN attribute getter.<p>
    *
    * @return the widget handle of the next sibling, in physical order, of the current browse
    *         column whether or not the column is visible.
    */
   public handle getNextColumn()
   {
      return getNextSibling();
   }

   /**
    * Gets the COLUMN-BGCOLOR writable attribute.
    *
    * @return   The current value of the COLUMN-BGCOLOR attribute.
    */
   public integer getBgColor()
   {
      int color = config.columnBgColor.color();
      return (color == -1) ? new integer() : new integer(color);
   }

   /**
    * Sets the COLUMN-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   public void setBgColor(integer bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setBgColor(color);
   }

   /**
    * Gets the LABEL-BGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-BGCOLOR attribute.
    */
   public integer getLabelBgColor()
   {
      int color = config.labelBgColor.color();
      return (color == -1) ? new integer() : new integer(color);
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   public void setLabelBgColor(integer bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setLabelBgColor(color);
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   public void setLabelBgColor(int bgcolor)
   {
      config.labelBgColor = new ColorSpec(bgcolor).convert();
      pushScreenDefinition();
   }

   /**
    * Set value of the SORT attribute.
    *
    * @param    sort
    *           <code>true</code> if the combo-box browse columns has sorted items.
    */
   public void setSort(boolean sort)
   {
      setSort(new logical(sort));
   }

   /**
    * Set value of the SORT attribute.
    *
    * @param    sort
    *           <code>true</code> if the combo-box browse columns has sorted items.
    */
   public void setSort(logical sort)
   {
      if (sort.isUnknown())
         return;

      config.sort = sort.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the SORT attribute.
    *
    * @return   <code>true</code> if the combo-box browse columns has sorted items.
    */
   public logical isSort()
   {
      return new logical(config.sort);
   }

   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int64 start, int64 end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int64 start, int end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int start, int64 end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public logical setSelection(int start, int end)
   {
      UnimplementedFeature.missing("SET-SELECTION method");
      return new logical();
   }
   
   /**
    * Obtains the index position of the cursor in the edited cell in this column.
    *
    * @return   The character index at which the cursor is placed.
    */
   public integer getCursorOffset()
   {
      // TODO implement
      return new integer();
   }

   /**
    * Sets the index position of the cursor in the edited cell in this column.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(double offset)
   {
      // TODO implement
   }

   /**
    * Sets the index position of the cursor in the edited cell in this column.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(NumberType offset)
   {
      // TODO implement
   }

   /**
    * Set value of the MAX-CHARS attribute. Applies only to combo-box browse columns.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   public void setMaxChars(double maxChars)
   {
      setMaxChars(new decimal(maxChars));
   }

   /**
    * Set value of the MAX-CHARS attribute. Applies only to combo-box browse columns.
    *
    * @param    maxChars
    *           Value of the attribute.
    */
   public void setMaxChars(NumberType maxChars)
   {
      if (maxChars.isUnknown())
         return;

      config.maxChars = maxChars.intValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the MAX-CHARS attribute. Applies only to combo-box browse columns.
    *
    * @return   value of the attribute.
    */
   public integer getMaxChars()
   {
      return new integer(config.maxChars);
   }

   /**
    * Set value of the AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    *
    * @param    autoZap
    *           Value of the attribute.
    */
   public void setAutoZap(logical autoZap)
   {
      // TODO implement
   }

   /**
    * Get value of the AUTO-ZAP attribute. Applies only to fill-in and combo-box browse columns.
    *
    * @return   value of the attribute.
    */
   public logical isAutoZap()
   {
      // TODO implement
      return new logical();
   }

   /**
    * Get the VISIBLE writable attribute.
    *
    * @return   The current value of VISIBLE attribute.
    */
   public boolean _isVisible()
   {
      return true;
   }

   /**
    * Gets the HIDDEN writable attribute.
    *
    * @return   The current value of the HIDDEN attribute.
    */
   public boolean _isHidden() 
   {
      return false;
   }
}
